"""
An object that can live on the dataset to facilitate data access.

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2015, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import weakref

from yt.extern.six import string_types
from yt.utilities.exceptions import YTDimensionalityError

class RegionExpression(object):
    _all_data = None
    def __init__(self, ds):
        self.ds = weakref.proxy(ds)

    @property
    def all_data(self):
        if self._all_data is None:
            self._all_data = self.ds.all_data()
        return self._all_data

    def __getitem__(self, item):
        # At first, we will only implement this as accepting a slice that is
        # (optionally) unitful corresponding to a specific set of coordinates
        # that result in a rectangular prism or a slice.
        if isinstance(item, string_types):
            # This is some field; we will instead pass this back to the
            # all_data object.
            return self.all_data[item]
        if isinstance(item, tuple) and isinstance(item[1], string_types):
            return self.all_data[item]
        if isinstance(item, slice):
            item = (item, item, item)
        if len(item) != self.ds.dimensionality:
            # Not the right specification, and we don't want to do anything
            # implicitly.
            raise YTDimensionalityError(len(item), self.ds.dimensionality)
        if self.ds.dimensionality != 3:
            # We'll pass on this for the time being.
            raise RuntimeError

        # OK, now we need to look at our slices.  How many are a specific
        # coordinate?
        
        if not all(isinstance(v, slice) for v in item):
            return self._create_slice(item)
        else:
            if all(s.start is s.stop is s.step is None for s in item):
                return self.all_data
            return self._create_region(item)
            
    def _spec_to_value(self, input_tuple):
        if not isinstance(input_tuple, tuple):
            # We now assume that it's in code_length
            return self.ds.quan(input_tuple, 'code_length')
        v, u = input_tuple
        value = self.ds.quan(v, u)
        return value

    def _create_slice(self, slice_tuple):
        axis = None
        new_slice = []
        for ax, v in enumerate(slice_tuple):
            if not isinstance(v, slice):
                if axis is not None: raise RuntimeError
                axis = ax
                coord = self._spec_to_value(v)
                new_slice.append(slice(None, None, None))
            else:
                new_slice.append(v)
        # This new slice doesn't need to be a tuple
        source = self._create_region(new_slice)
        sl = self.ds.slice(axis, coord, data_source = source)
        return sl

    def _slice_to_edges(self, ax, val):
        if val.start is None:
            l = self.ds.domain_left_edge[ax]
        else:
            l = self._spec_to_value(val.start)
        if val.stop is None:
            r = self.ds.domain_right_edge[ax]
        else:
            r = self._spec_to_value(val.stop)
        if r < l:
            raise RuntimeError
        return l, r

    def _create_region(self, bounds_tuple):
        left_edge = []
        right_edge = []
        dims = []
        for ax, b in enumerate(bounds_tuple):
            l, r = self._slice_to_edges(ax, b)
            left_edge.append(l)
            right_edge.append(r)
            dims.append(getattr(b.step, "imag", None))
        center = [ (cl + cr)/2.0 for cl, cr in zip(left_edge, right_edge)]
        if all(d is not None for d in dims):
            return self.ds.arbitrary_grid(left_edge, right_edge, dims)
        return self.ds.region(center, left_edge, right_edge)



import numpy as np

import yt
from yt.analysis_modules.level_sets.api import *

ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

data_source = ds.disk([0.5, 0.5, 0.5], [0., 0., 1.],
                      (8, 'kpc'), (1, 'kpc'))

# the field to be used for contouring
field = ("gas", "density")

# This is the multiplicative interval between contours.
step = 2.0

# Now we set some sane min/max values between which we want to find contours.
# This is how we tell the clump finder what to look for -- it won't look for
# contours connected below or above these threshold values.
c_min = 10**np.floor(np.log10(data_source[field]).min()  )
c_max = 10**np.floor(np.log10(data_source[field]).max()+1)

# Now find get our 'base' clump -- this one just covers the whole domain.
master_clump = Clump(data_source, field)

# Add a "validator" to weed out clumps with less than 20 cells.
# As many validators can be added as you want.
master_clump.add_validator("min_cells", 20)

# Begin clump finding.
find_clumps(master_clump, c_min, c_max, step)

# Write out the full clump hierarchy.
write_clump_index(master_clump, 0, "%s_clump_hierarchy.txt" % ds)

# Write out only the leaf nodes of the hierarchy.
write_clumps(master_clump,0, "%s_clumps.txt" % ds)

# We can traverse the clump hierarchy to get a list of all of the 'leaf' clumps
leaf_clumps = get_lowest_clumps(master_clump)

# If you'd like to visualize these clumps, a list of clumps can be supplied to
# the "clumps" callback on a plot.  First, we create a projection plot:
prj = yt.ProjectionPlot(ds, 2, field, center='c', width=(20,'kpc'))

# Next we annotate the plot with contours on the borders of the clumps
prj.annotate_clumps(leaf_clumps)

# Lastly, we write the plot to disk.
prj.save('clumps')



"""
Definitions for the lens code




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
cimport numpy as np
cimport cython
from .volume_container cimport VolumeContainer
from vec3_ops cimport dot, subtract, L2_norm, fma
from libc.math cimport exp, floor, log2, \
    fabs, atan, atan2, asin, cos, sin, sqrt, acos, M_PI
from yt.utilities.lib.fp_utils cimport imax, fmax, imin, fmin, iclip, fclip, i64clip

cdef extern from "platform_dep.h":
    long int lrint(double x) nogil

cdef extern from "limits.h":
    cdef int SHRT_MAX

cdef struct ImageContainer:
    np.float64_t[:,:,:] vp_pos
    np.float64_t[:,:,:] vp_dir
    np.float64_t *center
    np.float64_t[:,:,:] image
    np.float64_t[:,:] zbuffer
    np.int64_t[:,:] image_used
    np.int64_t[:,:] mesh_lines
    np.float64_t pdx, pdy
    np.float64_t bounds[4]
    np.float64_t[:,:] camera_data   # position, width, unit_vec[0,2]
    int nv[2]
    np.float64_t *x_vec
    np.float64_t *y_vec


ctypedef int calculate_extent_function(ImageContainer *image,
            VolumeContainer *vc, np.int64_t rv[4]) nogil except -1

ctypedef void generate_vector_info_function(ImageContainer *im,
            np.int64_t vi, np.int64_t vj,
            np.float64_t width[2],
            np.float64_t v_dir[3], np.float64_t v_pos[3]) nogil

cdef generate_vector_info_function generate_vector_info_plane_parallel
cdef generate_vector_info_function generate_vector_info_null
cdef calculate_extent_function calculate_extent_plane_parallel
cdef calculate_extent_function calculate_extent_perspective
cdef calculate_extent_function calculate_extent_null



"""
ExodusII-specific IO functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
from yt.utilities.io_handler import \
    BaseIOHandler
from yt.utilities.file_handler import \
    NetCDF4FileHandler


class IOHandlerExodusII(BaseIOHandler):
    _particle_reader = False
    _dataset_type = "exodus_ii"
    _INDEX_OFFSET = 1

    def __init__(self, ds):
        self.filename = ds.index_filename
        exodus_ii_handler = NetCDF4FileHandler(self.filename)
        self.handler = exodus_ii_handler.dataset
        super(IOHandlerExodusII, self).__init__(ds)
        self.node_fields = ds._get_nod_names()
        self.elem_fields = ds._get_elem_names()

    def _read_particle_coords(self, chunks, ptf):
        pass

    def _read_particle_fields(self, chunks, ptf, selector):
        pass

    def _read_fluid_selection(self, chunks, selector, fields, size):
        # This needs to allocate a set of arrays inside a dictionary, where the
        # keys are the (ftype, fname) tuples and the values are arrays that
        # have been masked using whatever selector method is appropriate.  The
        # dict gets returned at the end and it should be flat, with selected
        # data.  Note that if you're reading grid data, you might need to
        # special-case a grid selector object.
        chunks = list(chunks)
        rv = {}
        for field in fields:
            ftype, fname = field
            ci = self.handler.variables[ftype][:] - self._INDEX_OFFSET
            num_elem = ci.shape[0]
            if fname in self.node_fields:
                nodes_per_element = ci.shape[1]
                rv[field] = np.empty((num_elem, nodes_per_element), dtype="float64")
            elif fname in self.elem_fields:
                rv[field] = np.empty(num_elem, dtype="float64")
        for field in fields:
            ind = 0
            ftype, fname = field
            mesh_id = int(ftype[-1])
            chunk = chunks[mesh_id - 1]
            ci = self.handler.variables[ftype][:] - self._INDEX_OFFSET
            if fname in self.node_fields:
                field_ind = self.node_fields.index(fname)
                fdata = self.handler.variables['vals_nod_var%d' % (field_ind + 1)]
                data = fdata[self.ds.step][ci]
                for g in chunk.objs:
                    ind += g.select(selector, data, rv[field], ind)  # caches
            if fname in self.elem_fields:
                field_ind = self.elem_fields.index(fname)
                fdata = self.handler.variables['vals_elem_var%deb%s' %
                                               (field_ind + 1, mesh_id)][:]
                data = fdata[self.ds.step, :]
                for g in chunk.objs:
                    ind += g.select(selector, data, rv[field], ind)  # caches
        return rv

    def _read_chunk_data(self, chunk, fields):
        # This reads the data from a single chunk, and is only used for
        # caching.
        pass



"""
Tests for loading in-memory datasets



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import os
import shutil
import tempfile
import unittest

from yt.testing import assert_raises
from yt.convenience import load
from yt.utilities.exceptions import YTOutputNotIdentified

class TestEmptyLoad(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.curdir = os.getcwd()
        os.chdir(self.tmpdir)

        # create 0 byte file
        open("empty_file", "a")

        # create empty directory
        os.makedirs("empty_directory")

    def tearDown(self):
        os.chdir(self.curdir)
        shutil.rmtree(self.tmpdir)

    def test_load_empty_file(self):
        assert_raises(YTOutputNotIdentified, load, "not_a_file")
        assert_raises(YTOutputNotIdentified, load, "empty_file")
        assert_raises(YTOutputNotIdentified, load, "empty_directory")



from yt.testing import assert_equal
from yt.utilities.periodic_table import _elements, periodic_table

def test_element_accuracy():
    for num, w, name, sym in _elements:
        e0 = periodic_table[num]
        e1 = periodic_table[name]
        e2 = periodic_table[sym]
        # If num == -1, then we are in one of the things like Deuterium or El
        # that are not elements by themselves.
        if num == -1: e0 = e1
        yield assert_equal, id(e0), id(e1)
        yield assert_equal, id(e0), id(e2)
        yield assert_equal, e0.num, num
        yield assert_equal, e0.weight, w
        yield assert_equal, e0.name, name
        yield assert_equal, e0.symbol, sym



import numpy as np

from yt.testing import \
    fake_random_ds, \
    assert_equal, \
    assert_rel_equal

def setup():
    from yt.config import ytcfg
    ytcfg["yt","__withintesting"] = "True"

def test_extrema():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(16, nprocs = nprocs, fields = ("density",
                "velocity_x", "velocity_y", "velocity_z"))
        for sp in [ds.sphere("c", (0.25, 'unitary')), ds.r[0.5,:,:]]:
            mi, ma = sp.quantities["Extrema"]("density")
            yield assert_equal, mi, np.nanmin(sp["density"])
            yield assert_equal, ma, np.nanmax(sp["density"])
            dd = ds.all_data()
            mi, ma = dd.quantities["Extrema"]("density")
            yield assert_equal, mi, np.nanmin(dd["density"])
            yield assert_equal, ma, np.nanmax(dd["density"])
            sp = ds.sphere("max", (0.25, 'unitary'))
            yield assert_equal, np.any(np.isnan(sp["radial_velocity"])), False
            mi, ma = dd.quantities["Extrema"]("radial_velocity")
            yield assert_equal, mi, np.nanmin(dd["radial_velocity"])
            yield assert_equal, ma, np.nanmax(dd["radial_velocity"])

def test_average():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(16, nprocs = nprocs, fields = ("density",))
        for ad in [ds.all_data(), ds.r[0.5, :, :]]:
        
            my_mean = ad.quantities["WeightedAverageQuantity"]("density", "ones")
            yield assert_rel_equal, my_mean, ad["density"].mean(), 12

            my_mean = ad.quantities["WeightedAverageQuantity"]("density", "cell_mass")
            a_mean = (ad["density"] * ad["cell_mass"]).sum() / ad["cell_mass"].sum()
            yield assert_rel_equal, my_mean, a_mean, 12

def test_variance():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(16, nprocs = nprocs, fields = ("density", ))
        for ad in [ds.all_data(), ds.r[0.5, :, :]]:
        
            my_std, my_mean = ad.quantities["WeightedVariance"]("density", "ones")
            yield assert_rel_equal, my_mean, ad["density"].mean(), 12
            yield assert_rel_equal, my_std, ad["density"].std(), 12

            my_std, my_mean = ad.quantities["WeightedVariance"]("density", "cell_mass")        
            a_mean = (ad["density"] * ad["cell_mass"]).sum() / ad["cell_mass"].sum()
            yield assert_rel_equal, my_mean, a_mean, 12
            a_std = np.sqrt((ad["cell_mass"] * (ad["density"] - a_mean)**2).sum() / 
                            ad["cell_mass"].sum())
            yield assert_rel_equal, my_std, a_std, 12

def test_max_location():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(16, nprocs = nprocs, fields = ("density", ))
        for ad in [ds.all_data(), ds.r[0.5, :, :]]:

            mv, x, y, z = ad.quantities.max_location(("gas", "density"))

            yield assert_equal, mv, ad["density"].max()

            mi = np.argmax(ad["density"])

            yield assert_equal, ad["x"][mi], x
            yield assert_equal, ad["y"][mi], y
            yield assert_equal, ad["z"][mi], z

def test_min_location():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(16, nprocs = nprocs, fields = ("density", ))
        for ad in [ds.all_data(), ds.r[0.5, :, :]]:

            mv, x, y, z = ad.quantities.min_location(("gas", "density"))

            yield assert_equal, mv, ad["density"].min()

            mi = np.argmin(ad["density"])

            yield assert_equal, ad["x"][mi], x
            yield assert_equal, ad["y"][mi], y
            yield assert_equal, ad["z"][mi], z

def test_sample_at_min_field_values():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(16, nprocs = nprocs,
            fields = ("density", "temperature", "velocity_x"))
        for ad in [ds.all_data(), ds.r[0.5, :, :]]:

            mv, temp, vm = ad.quantities.sample_at_min_field_values(
                "density", ["temperature", "velocity_x"])

            yield assert_equal, mv, ad["density"].min()

            mi = np.argmin(ad["density"])

            yield assert_equal, ad["temperature"][mi], temp
            yield assert_equal, ad["velocity_x"][mi], vm

def test_sample_at_max_field_values():
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(16, nprocs = nprocs,
            fields = ("density", "temperature", "velocity_x"))
        for ad in [ds.all_data(), ds.r[0.5, :, :]]:

            mv, temp, vm = ad.quantities.sample_at_max_field_values(
                "density", ["temperature", "velocity_x"])

            yield assert_equal, mv, ad["density"].max()

            mi = np.argmax(ad["density"])

            yield assert_equal, ad["temperature"][mi], temp
            yield assert_equal, ad["velocity_x"][mi], vm

if __name__ == "__main__":
    for i in test_extrema():
        i[0](*i[1:])



from yt.utilities.answer_testing.framework import \
    requires_ds, \
    data_dir_load

enzotiny = "enzo_tiny_cosmology/DD0046/DD0046"
@requires_ds(enzotiny)
def test_simulated_halo_mass_function():
    ds = data_dir_load(enzotiny)
    for field_type in dir(ds.fields):
        ft = getattr(ds.fields, field_type)
        for field_name in dir(ft):
            f = getattr(ft, field_name)
            assert ((field_type, field_name) == f.name)



from yt.testing import assert_equal, fake_random_ds
from yt.units.unit_object import Unit
import os
import tempfile


def setup():
    from yt.config import ytcfg
    ytcfg["yt", "__withintesting"] = "True"


def teardown_func(fns):
    for fn in fns:
        try:
            os.remove(fn)
        except OSError:
            pass


def test_cutting_plane():
    fns = []
    for nprocs in [8, 1]:
        # We want to test both 1 proc and 8 procs, to make sure that
        # parallelism isn't broken
        ds = fake_random_ds(64, nprocs=nprocs)
        center = [0.5, 0.5, 0.5]
        normal = [1, 1, 1]
        cut = ds.cutting(normal, center)
        yield assert_equal, cut["ones"].sum(), cut["ones"].size
        yield assert_equal, cut["ones"].min(), 1.0
        yield assert_equal, cut["ones"].max(), 1.0
        pw = cut.to_pw(fields='density')
        for p in pw.plots.values():
            tmpfd, tmpname = tempfile.mkstemp(suffix='.png')
            os.close(tmpfd)
            p.save(name=tmpname)
            fns.append(tmpname)
        for width in [(1.0, 'unitary'), 1.0, ds.quan(0.5, 'code_length')]:
            frb = cut.to_frb(width, 64)
            for cut_field in ['ones', 'density']:
                fi = ds._get_field_info("unknown", cut_field)
                yield assert_equal, frb[cut_field].info['data_source'], \
                    cut.__str__()
                yield assert_equal, frb[cut_field].info['axis'], \
                    4
                yield assert_equal, frb[cut_field].info['field'], \
                    cut_field
                yield assert_equal, frb[cut_field].units, \
                    Unit(fi.units)
                yield assert_equal, frb[cut_field].info['xlim'], \
                    frb.bounds[:2]
                yield assert_equal, frb[cut_field].info['ylim'], \
                    frb.bounds[2:]
                yield assert_equal, frb[cut_field].info['length_to_cm'], \
                    ds.length_unit.in_cgs()
                yield assert_equal, frb[cut_field].info['center'], \
                    cut.center
    teardown_func(fns)



"""
Clump finding helper classes



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import copy
import numpy as np
import uuid

from yt.fields.derived_field import \
    ValidateSpatial
from yt.funcs import mylog, iterable
from yt.extern.six import string_types

from .clump_info_items import \
    clump_info_registry
from .clump_validators import \
    clump_validator_registry
from .contour_finder import \
    identify_contours

def add_contour_field(ds, contour_key):
    def _contours(field, data):
        fd = data.get_field_parameter("contour_slices_%s" % contour_key)
        vals = data["index", "ones"] * -1
        if fd is None or fd == 0.0:
            return vals
        for sl, v in fd.get(data.id, []):
            vals[sl] = v
        return vals

    ds.add_field(("index", "contours_%s" % contour_key),
                 function=_contours,
                 validators=[ValidateSpatial(0)],
                 take_log=False,
                 display_field=False,
                 units='')

class Clump(object):
    children = None
    def __init__(self, data, field, parent=None,
                 clump_info=None, validators=None):
        self.data = data
        self.field = field
        self.parent = parent
        self.quantities = data.quantities
        self.min_val = self.data[field].min()
        self.max_val = self.data[field].max()

        if parent is not None:
            self.data.parent = self.parent.data

        # List containing characteristics about clumps that are to be written 
        # out by the write routines.
        if clump_info is None:
            self.set_default_clump_info()
        else:
            # Clump info will act the same if add_info_item is called 
            # before or after clump finding.
            self.clump_info = copy.deepcopy(clump_info)

        if validators is None:
            validators = []
        self.validators = validators
        # Return value of validity function.
        self.valid = None

    def add_validator(self, validator, *args, **kwargs):
        """
        Add a validating function to determine whether the clump should 
        be kept.
        """
        callback = clump_validator_registry.find(validator, *args, **kwargs)
        self.validators.append(callback)
        if self.children is None: return
        for child in self.children:
            child.add_validator(validator)
        
    def add_info_item(self, info_item, *args, **kwargs):
        "Adds an entry to clump_info list and tells children to do the same."

        callback = clump_info_registry.find(info_item, *args, **kwargs)
        self.clump_info.append(callback)
        if self.children is None: return
        for child in self.children:
            child.add_info_item(info_item)

    def set_default_clump_info(self):
        "Defines default entries in the clump_info array."

        # add_info_item is recursive so this function does not need to be.
        self.clump_info = []

        self.add_info_item("total_cells")
        self.add_info_item("cell_mass")

        if any("jeans" in f for f in self.data.pf.field_list):
            self.add_info_item("mass_weighted_jeans_mass")
            self.add_info_item("volume_weighted_jeans_mass")

        self.add_info_item("max_grid_level")

        if any("number_density" in f for f in self.data.pf.field_list):
            self.add_info_item("min_number_density")
            self.add_info_item("max_number_density")

    def clear_clump_info(self):
        """
        Clears the clump_info array and passes the instruction to its 
        children.
        """

        self.clump_info = []
        if self.children is None: return
        for child in self.children:
            child.clear_clump_info()

    def write_info(self, level, f_ptr):
        "Writes information for clump using the list of items in clump_info."

        for item in self.clump_info:
            value = item(self)
            f_ptr.write("%s%s\n" % ('\t'*level, value))

    def find_children(self, min_val, max_val = None):
        if self.children is not None:
            mylog.info("Wiping out existing children clumps: %d.",
                       len(self.children))
        self.children = []
        if max_val is None: max_val = self.max_val
        nj, cids = identify_contours(self.data, self.field, min_val, max_val)
        # Here, cids is the set of slices and values, keyed by the
        # parent_grid_id, that defines the contours.  So we can figure out all
        # the unique values of the contours by examining the list here.
        unique_contours = set([])
        for sl_list in cids.values():
            for sl, ff in sl_list:
                unique_contours.update(np.unique(ff))
        contour_key = uuid.uuid4().hex
        base_object = getattr(self.data, 'base_object', self.data)
        add_contour_field(base_object.ds, contour_key)
        for cid in sorted(unique_contours):
            if cid == -1: continue
            new_clump = base_object.cut_region(
                    ["obj['contours_%s'] == %s" % (contour_key, cid)],
                    {('contour_slices_%s' % contour_key): cids})
            if new_clump["ones"].size == 0:
                # This is to skip possibly duplicate clumps.
                # Using "ones" here will speed things up.
                continue
            self.children.append(Clump(new_clump, self.field, parent=self,
                                       clump_info=self.clump_info,
                                       validators=self.validators))

    def pass_down(self,operation):
        """
        Performs an operation on a clump with an exec and passes the 
        instruction down to clump children.
        """

        # Call if callable, otherwise do an exec.
        if callable(operation):
            operation()
        else:
            exec(operation)

        if self.children is None: return
        for child in self.children:
            child.pass_down(operation)

    def _validate(self):
        "Apply all user specified validator functions."

        # Only call functions if not done already.
        if self.valid is not None:
            return self.valid

        self.valid = True
        for validator in self.validators:
            self.valid &= validator(self)
            if not self.valid:
                break

        return self.valid

    def __reduce__(self):
        return (_reconstruct_clump, 
                (self.parent, self.field, self.min_val, self.max_val,
                 self.valid, self.children, self.data, self.clump_info, 
                 self.function))

    def __getitem__(self,request):
        return self.data[request]

def _reconstruct_clump(parent, field, mi, ma, valid, children, 
                       data, clump_info, 
        function=None):
    obj = object.__new__(Clump)
    if iterable(parent):
        try:
            parent = parent[1]
        except KeyError:
            parent = parent
    if children is None: children = []
    obj.parent, obj.field, obj.min_val, obj.max_val, \
      obj.valid, obj.children, obj.clump_info, obj.function = \
        parent, field, mi, ma, valid, children, clump_info, function
    # Now we override, because the parent/child relationship seems a bit
    # unreliable in the unpickling
    for child in children: child.parent = obj
    obj.data = data[1] # Strip out the PF
    obj.quantities = obj.data.quantities
    if obj.parent is None: return (data[0], obj)
    return obj

def find_clumps(clump, min_val, max_val, d_clump):
    mylog.info("Finding clumps: min: %e, max: %e, step: %f" % 
               (min_val, max_val, d_clump))
    if min_val >= max_val: return
    clump.find_children(min_val)

    if (len(clump.children) == 1):
        find_clumps(clump, min_val*d_clump, max_val, d_clump)

    elif (len(clump.children) > 0):
        these_children = []
        mylog.info("Investigating %d children." % len(clump.children))
        for child in clump.children:
            find_clumps(child, min_val*d_clump, max_val, d_clump)
            if ((child.children is not None) and (len(child.children) > 0)):
                these_children.append(child)
            elif (child._validate()):
                these_children.append(child)
            else:
                mylog.info(("Eliminating invalid, childless clump with " +
                            "%d cells.") % len(child.data["ones"]))
        if (len(these_children) > 1):
            mylog.info("%d of %d children survived." %
                       (len(these_children),len(clump.children)))
            clump.children = these_children
        elif (len(these_children) == 1):
            mylog.info(("%d of %d children survived, linking its " +
                        "children to parent.") % 
                        (len(these_children),len(clump.children)))
            clump.children = these_children[0].children
        else:
            mylog.info("%d of %d children survived, erasing children." %
                       (len(these_children),len(clump.children)))
            clump.children = []

def get_lowest_clumps(clump, clump_list=None):
    "Return a list of all clumps at the bottom of the index."

    if clump_list is None: clump_list = []
    if clump.children is None or len(clump.children) == 0:
        clump_list.append(clump)
    if clump.children is not None and len(clump.children) > 0:
        for child in clump.children:
            get_lowest_clumps(child, clump_list=clump_list)

    return clump_list

def write_clump_index(clump, level, fh):
    top = False
    if isinstance(fh, string_types):
        fh = open(fh, "w")
        top = True
    for q in range(level):
        fh.write("\t")
    fh.write("Clump at level %d:\n" % level)
    clump.write_info(level, fh)
    fh.write("\n")
    fh.flush()
    if ((clump.children is not None) and (len(clump.children) > 0)):
        for child in clump.children:
            write_clump_index(child, (level+1), fh)
    if top:
        fh.close()

def write_clumps(clump, level, fh):
    top = False
    if isinstance(fh, string_types):
        fh = open(fh, "w")
        top = True
    if ((clump.children is None) or (len(clump.children) == 0)):
        fh.write("%sClump:\n" % ("\t"*level))
        clump.write_info(level, fh)
        fh.write("\n")
        fh.flush()
    if ((clump.children is not None) and (len(clump.children) > 0)):
        for child in clump.children:
            write_clumps(child, 0, fh)
    if top:
        fh.close()



import numpy as np

from yt.data_objects.profiles import \
    Profile1D, \
    Profile2D, \
    Profile3D, \
    create_profile
from yt.testing import \
    fake_random_ds, \
    assert_equal, \
    assert_raises, \
    assert_rel_equal
from yt.utilities.exceptions import \
    YTIllDefinedProfile
from yt.visualization.profile_plotter import \
    ProfilePlot, \
    PhasePlot

_fields = ("density", "temperature", "dinosaurs", "tribbles")
_units = ("g/cm**3", "K", "dyne", "erg")


def test_profiles():
    ds = fake_random_ds(64, nprocs = 8, fields = _fields, units = _units)
    nv = ds.domain_dimensions.prod()
    dd = ds.all_data()
    (rmi, rma), (tmi, tma), (dmi, dma) = dd.quantities["Extrema"](
        ["density", "temperature", "dinosaurs"])
    rt, tt, dt = dd.quantities["TotalQuantity"](
        ["density", "temperature", "dinosaurs"])

    e1, e2 = 0.9, 1.1
    for nb in [8, 16, 32, 64]:
        for input_units in ['mks', 'cgs']:
            for ex in [rmi, rma, tmi, tma, dmi, dma]:
                getattr(ex, 'convert_to_%s' % input_units)()
            # We log all the fields or don't log 'em all.  No need to do them
            # individually.
            for lf in [True, False]:
                direct_profile = Profile1D(
                    dd, "density", nb, rmi*e1, rma*e2, lf, weight_field=None)
                direct_profile.add_fields(["ones", "temperature"])

                indirect_profile_s = create_profile(
                    dd, "density", ["ones", "temperature"], n_bins=nb,
                    extrema={'density': (rmi*e1, rma*e2)}, logs={'density': lf},
                    weight_field=None)

                indirect_profile_t = create_profile(
                    dd, ("gas", "density"),
                    [("index", "ones"), ("gas", "temperature")], n_bins=nb,
                    extrema={'density': (rmi*e1, rma*e2)}, logs={'density': lf},
                    weight_field=None)

                for p1d in [direct_profile, indirect_profile_s,
                            indirect_profile_t]:
                    yield assert_equal, p1d["index", "ones"].sum(), nv
                    yield assert_rel_equal, tt, \
                        p1d["gas", "temperature"].sum(), 7

                p2d = Profile2D(
                    dd,
                    "density",     nb, rmi*e1, rma*e2, lf,
                    "temperature", nb, tmi*e1, tma*e2, lf,
                    weight_field=None)
                p2d.add_fields(["ones", "temperature"])
                yield assert_equal, p2d["ones"].sum(), nv
                yield assert_rel_equal, tt, p2d["temperature"].sum(), 7

                p3d = Profile3D(
                    dd,
                    "density",     nb, rmi*e1, rma*e2, lf,
                    "temperature", nb, tmi*e1, tma*e2, lf,
                    "dinosaurs",   nb, dmi*e1, dma*e2, lf,
                    weight_field=None)
                p3d.add_fields(["ones", "temperature"])
                yield assert_equal, p3d["ones"].sum(), nv
                yield assert_rel_equal, tt, p3d["temperature"].sum(), 7

        p1d = Profile1D(dd, "x", nb, 0.0, 1.0, False,
                        weight_field = None)
        p1d.add_fields("ones")
        av = nv / nb
        yield assert_equal, p1d["ones"], np.ones(nb)*av

        # We re-bin ones with a weight now
        p1d = Profile1D(dd, "x", nb, 0.0, 1.0, False,
                        weight_field = "temperature")
        p1d.add_fields(["ones"])
        yield assert_equal, p1d["ones"], np.ones(nb)

        # Verify we can access "ones" after adding a new field
        # See issue 988
        p1d.add_fields(["density"])
        yield assert_equal, p1d["ones"], np.ones(nb)

        p2d = Profile2D(dd, "x", nb, 0.0, 1.0, False,
                            "y", nb, 0.0, 1.0, False,
                            weight_field = None)
        p2d.add_fields("ones")
        av = nv / nb**2
        yield assert_equal, p2d["ones"], np.ones((nb, nb))*av

        # We re-bin ones with a weight now
        p2d = Profile2D(dd, "x", nb, 0.0, 1.0, False,
                            "y", nb, 0.0, 1.0, False,
                            weight_field = "temperature")
        p2d.add_fields(["ones"])
        yield assert_equal, p2d["ones"], np.ones((nb, nb))

        p3d = Profile3D(dd, "x", nb, 0.0, 1.0, False,
                            "y", nb, 0.0, 1.0, False,
                            "z", nb, 0.0, 1.0, False,
                            weight_field = None)
        p3d.add_fields("ones")
        av = nv / nb**3
        yield assert_equal, p3d["ones"], np.ones((nb, nb, nb))*av

        # We re-bin ones with a weight now
        p3d = Profile3D(dd, "x", nb, 0.0, 1.0, False,
                            "y", nb, 0.0, 1.0, False,
                            "z", nb, 0.0, 1.0, False,
                            weight_field = "temperature")
        p3d.add_fields(["ones"])
        yield assert_equal, p3d["ones"], np.ones((nb,nb,nb))

extrema_s = {'particle_position_x': (0, 1)}
logs_s = {'particle_position_x': False}

extrema_t = {('all', 'particle_position_x'): (0, 1)}
logs_t = {('all', 'particle_position_x'): False}

def test_particle_profiles():
    for nproc in [1, 2, 4, 8]:
        ds = fake_random_ds(32, nprocs=nproc, particles = 32**3)
        dd = ds.all_data()

        p1d = Profile1D(dd, "particle_position_x", 128,
                        0.0, 1.0, False, weight_field = None)
        p1d.add_fields(["particle_ones"])
        yield assert_equal, p1d["particle_ones"].sum(), 32**3

        p1d = create_profile(dd, ["particle_position_x"], ["particle_ones"],
                             weight_field=None, n_bins=128, extrema=extrema_s,
                             logs=logs_s)
        yield assert_equal, p1d["particle_ones"].sum(), 32**3

        p1d = create_profile(dd,
                             [("all", "particle_position_x")],
                             [("all", "particle_ones")],
                             weight_field=None, n_bins=128, extrema=extrema_t,
                             logs=logs_t)
        yield assert_equal, p1d["particle_ones"].sum(), 32**3

        p2d = Profile2D(dd, "particle_position_x", 128, 0.0, 1.0, False,
                            "particle_position_y", 128, 0.0, 1.0, False,
                        weight_field = None)
        p2d.add_fields(["particle_ones"])
        yield assert_equal, p2d["particle_ones"].sum(), 32**3

        p3d = Profile3D(dd, "particle_position_x", 128, 0.0, 1.0, False,
                            "particle_position_y", 128, 0.0, 1.0, False,
                            "particle_position_z", 128, 0.0, 1.0, False,
                        weight_field = None)
        p3d.add_fields(["particle_ones"])
        yield assert_equal, p3d["particle_ones"].sum(), 32**3

def test_mixed_particle_mesh_profiles():
    ds = fake_random_ds(32, particles=10)
    ad = ds.all_data()
    assert_raises(
        YTIllDefinedProfile, ProfilePlot, ad, 'radius', 'particle_mass')
    assert_raises(
        YTIllDefinedProfile, ProfilePlot, ad, 'radius',
        ['particle_mass', 'particle_ones'])
    assert_raises(
        YTIllDefinedProfile, ProfilePlot, ad, 'radius',
        ['particle_mass', 'ones'])
    assert_raises(
        YTIllDefinedProfile, ProfilePlot, ad, 'particle_radius', 'particle_mass',
        'cell_mass')
    assert_raises(
        YTIllDefinedProfile, ProfilePlot, ad, 'radius', 'cell_mass',
        'particle_ones')

    assert_raises(
        YTIllDefinedProfile, PhasePlot, ad, 'radius', 'particle_mass',
        'velocity_x')
    assert_raises(
        YTIllDefinedProfile, PhasePlot, ad, 'particle_radius', 'particle_mass',
        'cell_mass')
    assert_raises(
        YTIllDefinedProfile, PhasePlot, ad, 'radius', 'cell_mass',
        'particle_ones')
    assert_raises(
        YTIllDefinedProfile, PhasePlot, ad, 'particle_radius', 'particle_mass',
        'particle_ones')



import numpy as np

from yt.testing import \
    fake_random_ds, \
    assert_array_less

def setup():
    from yt.config import ytcfg
    ytcfg["yt","loglevel"] = "50"
    ytcfg["yt","__withintesting"] = "True"

def _difference(x1, x2, dw):
    rel = x1 - x2
    rel[rel >  dw/2.0] -= dw
    rel[rel < -dw/2.0] += dw
    return rel

def test_ellipsoid():
    # We decompose in different ways
    cs = [
          np.array([0.5, 0.5, 0.5]),
          np.array([0.1, 0.2, 0.3]),
          np.array([0.8, 0.8, 0.8])
          ]
    np.random.seed(int(0x4d3d3d3))
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(64, nprocs = nprocs)
        DW = ds.domain_right_edge - ds.domain_left_edge
        min_dx = 2.0/ds.domain_dimensions
        ABC = np.random.random((3, 12)) * 0.1
        e0s = np.random.random((3, 12))
        tilts = np.random.random(12)
        ABC[:,0] = 0.1
        for i in range(12):
            for c in cs:
                A, B, C = reversed(sorted(ABC[:,i]))
                A = max(A, min_dx[0])
                B = max(B, min_dx[1])
                C = max(C, min_dx[2])
                e0 = e0s[:,i]
                tilt = tilts[i]
                ell = ds.ellipsoid(c, A, B, C, e0, tilt)
                yield assert_array_less, ell["radius"], A
                p = np.array([ell[ax] for ax in 'xyz'])
                dot_evec = [np.zeros_like(ell["radius"]) for i in range(3)]
                vecs = [ell._e0, ell._e1, ell._e2]
                mags = [ell._A, ell._B, ell._C]
                my_c = np.array([c]*p.shape[1]).transpose()
                dot_evec = [de.to_ndarray() for de in dot_evec]
                mags = [m.to_ndarray() for m in mags]
                for ax_i in range(3):
                    dist = _difference(p[ax_i,:], my_c[ax_i,:], DW[ax_i])
                    for ax_j in range(3):
                        dot_evec[ax_j] += dist * vecs[ax_j][ax_i]
                dist = 0
                for ax_i in range(3):
                    dist += dot_evec[ax_i]**2.0 / mags[ax_i]**2.0
                yield assert_array_less, dist, 1.0



import string
from itertools import takewhile
from netCDF4 import Dataset
import numpy as np
from yt.config import ytcfg
import os


def sanitize_string(s):
    s = "".join(_ for _ in takewhile(lambda a: a in string.printable, s))
    return s


def get_data(fn):
    try:
        f = Dataset(fn)
    except RuntimeError:
        f = Dataset(os.path.join(ytcfg.get("yt", "test_data_dir"), fn))
    fvars = f.variables
    # Is this correct?
    etypes = fvars["eb_status"][:]
    nelem = etypes.shape[0]
    varnames = [sanitize_string(v.tostring()) for v in
                fvars["name_elem_var"][:]]
    nodnames = [sanitize_string(v.tostring()) for v in
                fvars["name_nod_var"][:]]
    coord = np.array([fvars["coord%s" % ax][:]
                     for ax in 'xyz']).transpose().copy()
    coords = []
    connects = []
    data = []
    for i in range(nelem):
        connects.append(fvars["connect%s" % (i+1)][:].astype("i8"))
        ci = connects[-1]
        coords.append(coord)  # Same for all
        vals = {}
        for j, v in enumerate(varnames):
            values = fvars["vals_elem_var%seb%s" % (j+1, i+1)][:]
            vals['gas', v] = values.astype("f8")[-1, :]
        for j, v in enumerate(nodnames):
            # We want just for this set of nodes all the node variables
            # Use (ci - 1) to get these values
            values = fvars["vals_nod_var%s" % (j+1)][:]
            vals['gas', v] = values.astype("f8")[-1, ci - 1, ...]
        data.append(vals)
    return coords, connects, data



"""
BoxLib code fields

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import string
import re

from yt.utilities.physical_constants import \
    boltzmann_constant_cgs, amu_cgs
from yt.fields.field_info_container import \
    FieldInfoContainer

rho_units = "code_mass / code_length**3"
mom_units = "code_mass / (code_time * code_length**2)"
eden_units = "code_mass / (code_time**2 * code_length)" # erg / cm^3

spec_finder = re.compile(r'.*\((\D*)(\d*)\).*')


def _thermal_energy_density(field, data):
    # What we've got here is UEINT:
    # u here is velocity
    # E is energy density from the file
    #   rho e = rho E - rho * u * u / 2
    ke = 0.5 * ( data["momentum_x"]**2
               + data["momentum_y"]**2
               + data["momentum_z"]**2) / data["density"]
    return data["eden"] - ke


def _thermal_energy(field, data):
    # This is little e, so we take thermal_energy_density and divide by density
    return data["thermal_energy_density"] / data["density"]


def _temperature(field, data):
    mu = data.ds.parameters["mu"]
    gamma = data.ds.parameters["gamma"]
    tr = data["thermal_energy_density"] / data["density"]
    tr *= mu * amu_cgs / boltzmann_constant_cgs
    tr *= (gamma - 1.0)
    return tr


class BoxlibFieldInfo(FieldInfoContainer):
    known_other_fields = (
        ("density", (rho_units, ["density"], None)),
        ("eden", (eden_units, ["energy_density"], None)),
        ("xmom", (mom_units, ["momentum_x"], None)),
        ("ymom", (mom_units, ["momentum_y"], None)),
        ("zmom", (mom_units, ["momentum_z"], None)),
        ("temperature", ("K", ["temperature"], None)),
        ("Temp", ("K", ["temperature"], None)),
        ("x_velocity", ("cm/s", ["velocity_x"], None)),
        ("y_velocity", ("cm/s", ["velocity_y"], None)),
        ("z_velocity", ("cm/s", ["velocity_z"], None)),
        ("xvel", ("cm/s", ["velocity_x"], None)),
        ("yvel", ("cm/s", ["velocity_y"], None)),
        ("zvel", ("cm/s", ["velocity_z"], None)),
    )

    known_particle_fields = (
        ("particle_mass", ("code_mass", [], None)),
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
        ("particle_momentum_x", ("code_mass*code_length/code_time", [], None)),
        ("particle_momentum_y", ("code_mass*code_length/code_time", [], None)),
        ("particle_momentum_z", ("code_mass*code_length/code_time", [], None)),
        # Note that these are *internal* agmomen
        ("particle_angmomen_x", ("code_length**2/code_time", [], None)),
        ("particle_angmomen_y", ("code_length**2/code_time", [], None)),
        ("particle_angmomen_z", ("code_length**2/code_time", [], None)),
        ("particle_id", ("", ["particle_index"], None)),
        ("particle_mdot", ("code_mass/code_time", [], None)),
        # "mlast",
        # "r",
        # "mdeut",
        # "n",
        # "burnstate",
        # "luminosity",
    )

    def setup_particle_fields(self, ptype):

        def _get_vel(axis):
            def velocity(field, data):
                return data["particle_momentum_%s" % axis]/data["particle_mass"]
            return velocity

        for ax in 'xyz':
            self.add_field((ptype, "particle_velocity_%s" % ax), 
                           function=_get_vel(ax),
                           particle_type=True,
                           units="code_length/code_time")

        super(BoxlibFieldInfo, self).setup_particle_fields(ptype)

    def setup_fluid_fields(self):
        unit_system = self.ds.unit_system
        # Now, let's figure out what fields are included.
        if any(f[1] == "xmom" for f in self.field_list):
            self.setup_momentum_to_velocity()
        elif any(f[1] == "xvel" for f in self.field_list):
            self.setup_velocity_to_momentum()
        self.add_field(("gas", "thermal_energy"),
                       function=_thermal_energy,
                       units=unit_system["specific_energy"])
        self.add_field(("gas", "thermal_energy_density"),
                       function=_thermal_energy_density,
                       units=unit_system["pressure"])
        if ("gas", "temperature") not in self.field_aliases:
            self.add_field(("gas", "temperature"),
                           function=_temperature,
                           units=unit_system["temperature"])

    def setup_momentum_to_velocity(self):
        def _get_vel(axis):
            def velocity(field, data):
                return data["%smom" % axis]/data["density"]
            return velocity
        for ax in 'xyz':
            self.add_field(("gas", "velocity_%s" % ax),
                           function=_get_vel(ax),
                           units=self.ds.unit_system["velocity"])

    def setup_velocity_to_momentum(self):
        def _get_mom(axis):
            def momentum(field, data):
                return data["%svel" % axis]*data["density"]
            return momentum
        for ax in 'xyz':
            self.add_field(("gas", "momentum_%s" % ax),
                           function=_get_mom(ax),
                           units=mom_units)


class CastroFieldInfo(FieldInfoContainer):

    known_other_fields = (
        ("density", ("g/cm**3", ["density"], r"\rho")),
        ("xmom", ("g/(cm**2 * s)", ["momentum_x"], r"\rho u")),
        ("ymom", ("g/(cm**2 * s)", ["momentum_y"], r"\rho v")),
        ("zmom", ("g/(cm**2 * s)", ["momentum_z"], r"\rho w")),
        # velocity components are not always present
        ("x_velocity", ("cm/s", ["velocity_x"], r"u")),
        ("y_velocity", ("cm/s", ["velocity_y"], r"v")),
        ("z_velocity", ("cm/s", ["velocity_z"], r"w")),
        ("rho_E", ("erg/cm**3", ["energy_density"], r"\rho E")),
        # internal energy density (not just thermal)
        ("rho_e", ("erg/cm**3", [], r"\rho e")),
        ("Temp", ("K", ["temperature"], r"T")),
        ("grav_x", ("cm/s**2", [],
                    r"\mathbf{g} \cdot \mathbf{e}_x")),
        ("grav_y", ("cm/s**2", [],
                    r"\mathbf{g} \cdot \mathbf{e}_y")),
        ("grav_z", ("cm/s**2", [],
                    r"\mathbf{g} \cdot \mathbf{e}_z")),
        ("pressure", ("dyne/cm**2", [], r"p")),
        ("kineng", ("erg/cm**3", ["kinetic_energy"], r"\frac{1}{2}\rho|\mathbf{U}|^2")),
        ("soundspeed", ("cm/s", ["sound_speed"], "Sound Speed")),
        ("Machnumber", ("", ["mach_number"], "Mach Number")),
        ("entropy", ("erg/(g*K)", ["entropy"], r"s")),
        ("magvort", ("1/s", ["vorticity_magnitude"],
                     r"|\nabla \times \mathbf{U}|")),
        ("divu", ("1/s", ["velocity_divergence"], r"\nabla \cdot \mathbf{U}")),
        ("eint_E", ("erg/g", [], r"e(E,U)")),
        ("eint_e", ("erg/g", [], r"e")),
        ("magvel", ("cm/s", ["velocity_magnitude"], r"|\mathbf{U}|")),
        ("radvel", ("cm/s", ["radial_velocity"],
                    r"\mathbf{U} \cdot \mathbf{e}_r")),
        ("magmom", ("g*cm/s", ["momentum_magnitude"], r"\rho |\mathbf{U}|")),
        ("maggrav", ("cm/s**2", [], r"|\mathbf{g}|")),
        ("phiGrav", ("erg/g", [], r"\Phi")),
        ("enuc", ("erg/(g*s)", [], r"\dot{e}_{\rm{nuc}}")),
        ("rho_enuc", ("erg/(cm**3*s)", [], r"\rho \dot{e}_{\rm{nuc}}")),
        ("angular_momentum_x", ("g/(cm*s)", [], r"\ell_x")),
        ("angular_momentum_y", ("g/(cm*s)", [], r"\ell_y")),
        ("angular_momentum_z", ("g/(cm*s)", [], r"\ell_z")),
        ("phiRot", ("erg/g", [], r"\Phi_{\rm{rot}}")),
        ("rot_x", ("cm/s**2", [], r"\mathbf{f}_{\rm{rot}} \cdot \mathbf{e}_x")),
        ("rot_y", ("cm/s**2", [], r"\mathbf{f}_{\rm{rot}} \cdot \mathbf{e}_y")),
        ("rot_z", ("cm/s**2", [], r"\mathbf{f}_{\rm{rot}} \cdot \mathbf{e}_z")),
    )

    def setup_fluid_fields(self):
        # add X's
        for _, field in self.ds.field_list:
            if field.startswith("X("):
                # We have a fraction
                nice_name, tex_label = _nice_species_name(field)
                self.alias(("gas", "%s_fraction" % nice_name),
                           ("boxlib", field),
                           units="")
                func = _create_density_func(("gas", "%s_fraction" % nice_name))
                self.add_field(name=("gas", "%s_density" % nice_name),
                               function = func,
                               units = self.ds.unit_system["density"])
                # We know this will either have one letter, or two.
                if field[3] in string.ascii_letters:
                    element, weight = field[2:4], field[4:-1]
                else:
                    element, weight = field[2:3], field[3:-1]  # NOQA

                # Here we can, later, add number density
                # right now element and weight inferred above are unused


class MaestroFieldInfo(FieldInfoContainer):

    known_other_fields = (
        ("density", ("g/cm**3", ["density"], None)),
        ("x_vel", ("cm/s", ["velocity_x"], r"\tilde{u}")),
        ("y_vel", ("cm/s", ["velocity_y"], r"\tilde{v}")),
        ("z_vel", ("cm/s", ["velocity_z"], r"\tilde{w}")),
        ("magvel", ("cm/s", ["velocity_magnitude"],
                    r"|\tilde{\mathbf{U}} + w_0 \mathbf{e}_r|")),
        ("radial_velocity", ("cm/s", ["radial_velocity"], r"\mathbf{U}\cdot \mathbf{e}_r")),
        ("circum_velocity", ("cm/s", ["tangential_velocity"], r"U - U\cdot e_r")),
        ("tfromp", ("K", [], "T(\\rho,p,X)")),
        ("tfromh", ("K", [], "T(\\rho,h,X)")),
        ("Machnumber", ("", ["mach_number"], "M")),
        ("S", ("1/s", [], None)),
        ("ad_excess", ("", [], r"\nabla - \nabla_\mathrm{ad}")),
        ("deltaT", ("", [], "[T(\\rho,h,X) - T(\\rho,p,X)]/T(\\rho,h,X)")),
        ("deltagamma", ("", [], "\Gamma_1 - \overline{\Gamma_1}")),
        ("deltap", ("", [], "[p(\\rho,h,X) - p_0] / p_0")),
        ("divw0", ("1/s", [], r"\nabla \cdot \mathbf{w}_0")),
        # Specific entropy
        ("entropy", ("erg/(g*K)", ["entropy"], "s")),
        ("entropypert", ("", [], "[s - \overline{s}] / \overline{s}")),
        ("enucdot", ("erg/(g*s)", [], "\dot{\epsilon}_{nuc}")),
        ("Hext", ("erg/(g*s)", [], "H_{ext}")),
        # Perturbational pressure grad
        ("gpi_x", ("dyne/cm**3", [], r"\left(\nabla\pi\right)_x")),
        ("gpi_y", ("dyne/cm**3", [], r"\left(\nabla\pi\right)_y")),
        ("gpi_z", ("dyne/cm**3", [], r"\left(\nabla\pi\right)_z")),
        ("h", ("erg/g", [], "h")),
        ("h0", ("erg/g", [], "h_0")),
        # Momentum cannot be computed because we need to include base and
        # full state.
        ("momentum", ("g*cm/s", ["momentum_magnitude"], r"\rho |\mathbf{U}|")),
        ("p0", ("erg/cm**3", [], "p_0")),
        ("p0pluspi", ("erg/cm**3", [], "p_0 + \pi")),
        ("pi", ("erg/cm**3", [], "\pi")),
        ("pioverp0", ("", [], "\pi/p_0")),
        # Base state density
        ("rho0", ("g/cm**3", [], "\\rho_0")),
        ("rhoh", ("erg/cm**3", ["enthalpy_density"], "(\\rho h)")),
        # Base state enthalpy density
        ("rhoh0", ("erg/cm**3", [], "(\\rho h)_0")),
        ("rhohpert", ("erg/cm**3", [], "(\\rho h)^\prime")),
        ("rhopert", ("g/cm**3", [], "\\rho^\prime")),
        ("soundspeed", ("cm/s", ["sound_speed"], None)),
        ("sponge", ("", [], None)),
        ("tpert", ("K", [], "T - \overline{T}")),
        # Again, base state -- so we can't compute ourselves.
        ("vort", ("1/s", ["vorticity_magnitude"], r"|\nabla\times\tilde{U}|")),
        # Base state
        ("w0_x", ("cm/s", [], "(w_0)_x")),
        ("w0_y", ("cm/s", [], "(w_0)_y")),
        ("w0_z", ("cm/s", [], "(w_0)_z")),
    )

    def setup_fluid_fields(self):
        unit_system = self.ds.unit_system
        # pick the correct temperature field
        if self.ds.parameters["use_tfromp"]:
            self.alias(("gas", "temperature"), ("boxlib", "tfromp"),
                       units=unit_system["temperature"])
        else:
            self.alias(("gas", "temperature"), ("boxlib", "tfromh"),
                       units=unit_system["temperature"])

        # Add X's and omegadots, units of 1/s
        for _, field in self.ds.field_list:
            if field.startswith("X("):
                # We have a mass fraction
                nice_name, tex_label = _nice_species_name(field)
                # Overwrite field to use nicer tex_label display_name
                self.add_output_field(("boxlib", field),
                                      units="",
                                      display_name=tex_label)
                self.alias(("gas", "%s_fraction" % nice_name),
                           ("boxlib", field),
                           units="")
                func = _create_density_func(("gas", "%s_fraction" % nice_name))
                self.add_field(name=("gas", "%s_density" % nice_name),
                               function=func,
                               units=unit_system["density"],
                               display_name=r'\rho %s' % tex_label)

                # Most of the time our species will be of the form
                # element name + atomic weight (e.g. C12), but
                # sometimes we make up descriptive names (e.g. ash)
                if any(char.isdigit() for char in field):
                    # We know this will either have one letter, or two.
                    if field[3] in string.ascii_letters:
                        element, weight = field[2:4], field[4:-1]
                    else:
                        element, weight = field[2:3], field[3:-1]  # NOQA
                    weight = int(weight)

                # Here we can, later, add number density using 'element' and
                # 'weight' inferred above

            elif field.startswith("omegadot("):
                nice_name, tex_label = _nice_species_name(field)
                display_name = r'\dot{\omega}\left[%s\right]' % tex_label
                # Overwrite field to use nicer tex_label'ed display_name
                self.add_output_field(("boxlib", field), units=unit_system["frequency"],
                                      display_name=display_name)
                self.alias(("gas", "%s_creation_rate" % nice_name),
                           ("boxlib", field), units=unit_system["frequency"])


def _nice_species_name(field):
    spec_match = spec_finder.search(field)
    nice_name = ''.join(spec_match.groups())
    # if the species field is a descriptive name, then the match
    # on the integer will be blank
    # modify the tex string in this case to remove spurious tex spacing
    lab = r"X\left(^{%s}%s\right)"
    if spec_match.groups()[-1] == "":
        lab = r"X\left(%s%s\right)"
    tex_label = lab % spec_match.groups()[::-1]
    return nice_name, tex_label


def _create_density_func(field_name):
    def _func(field, data):
        return data[field_name] * data["gas", "density"]
    return _func



"""
GAMER-specific data structures



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import os
import stat
import numpy as np
import weakref

from yt.funcs import \
    mylog, \
    setdefaultattr
from yt.data_objects.grid_patch import \
    AMRGridPatch
from yt.geometry.grid_geometry_handler import \
    GridIndex
from yt.data_objects.static_output import \
    Dataset
from yt.utilities.file_handler import \
    HDF5FileHandler
from .fields import GAMERFieldInfo
from yt.testing import assert_equal



class GAMERGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, id, index, level):
        AMRGridPatch.__init__(self, id,
                              filename = index.index_filename,
                              index    = index)
        self.Parent   = None    # do NOT initialize Parent as []
        self.Children = []
        self.Level    = level

    def __repr__(self):
        return 'GAMERGrid_%09i (dimension = %s)' % (self.id, self.ActiveDimensions)


class GAMERHierarchy(GridIndex):
    grid                 = GAMERGrid
    _preload_implemented = True # since gamer defines "_read_chunk_data" in io.py
    
    def __init__(self, ds, dataset_type = 'gamer'):
        self.dataset_type     = dataset_type
        self.dataset          = weakref.proxy(ds)
        self.index_filename   = self.dataset.parameter_filename
        self.directory        = os.path.dirname(self.index_filename)
        self._handle          = ds._handle
        self.float_type       = 'float64' # fixed even when FLOAT8 is off
        self._particle_handle = ds._particle_handle
        GridIndex.__init__(self, ds, dataset_type)

    def _detect_output_fields(self):
        # find all field names in the current dataset
        self.field_list = [ ('gamer', v) for v in self._handle['Data'].keys() ]
    
    def _count_grids(self):
        # count the total number of patches at all levels  
        self.num_grids = self.dataset.parameters['NPatch'].sum()
        
    def _parse_index(self):
        parameters       = self.dataset.parameters
        gid0             = 0
        grid_corner      = self._handle['Tree/Corner'].value
        convert2physical = self._handle['Tree/Corner'].attrs['Cvt2Phy']

        self.grid_dimensions    [:] = parameters['PatchSize']
        self.grid_particle_count[:] = 0

        for lv in range(0, parameters['NLevel']):
            num_grids_level = parameters['NPatch'][lv]
            if num_grids_level == 0: break

            patch_scale = parameters['PatchSize']*parameters['CellScale'][lv]

            # set the level and edge of each grid
            # (left/right_edge are YT arrays in code units)
            self.grid_levels.flat[ gid0:gid0 + num_grids_level ] = lv
            self.grid_left_edge[ gid0:gid0 + num_grids_level ] \
                = grid_corner[ gid0:gid0 + num_grids_level ]*convert2physical
            self.grid_right_edge[ gid0:gid0 + num_grids_level ] \
                = (grid_corner[ gid0:gid0 + num_grids_level ] + patch_scale)*convert2physical

            gid0 += num_grids_level

        # allocate all grid objects
        self.grids = np.empty(self.num_grids, dtype='object')
        for i in range(self.num_grids):
            self.grids[i] = self.grid(i, self, self.grid_levels.flat[i])

        # maximum level with patches (which can be lower than MAX_LEVEL)
        self.max_level = self.grid_levels.max()
        
    def _populate_grid_objects(self):
        son_list = self._handle["Tree/Son"].value

        for gid in range(self.num_grids):
            grid     = self.grids.flat[gid]
            son_gid0 = son_list[gid]

            # set up the parent-children relationship
            if son_gid0 >= 0:
                grid.Children = [ self.grids.flat[son_gid0+s] for s in range(8) ]

            for son_grid in grid.Children: son_grid.Parent = grid

            # set up other grid attributes
            grid._prepare_grid()
            grid._setup_dx()

        # validate the parent-children relationship in the debug mode
        if self.dataset._debug:
            self._validate_parent_children_relasionship()

    # for _debug mode only
    def _validate_parent_children_relasionship(self):
        mylog.info('Validating the parent-children relationship ...')

        father_list = self._handle["Tree/Father"].value

        for grid in self.grids:
            # parent->children == itself
            if grid.Parent is not None:
                assert grid.Parent.Children[0+grid.id%8] is grid, \
                       'Grid %d, Parent %d, Parent->Children %d' % \
                       (grid.id, grid.Parent.id, grid.Parent.Children[0].id)

            # children->parent == itself
            for c in grid.Children:
                assert c.Parent is grid, \
                       'Grid %d, Children %d, Children->Parent %d' % \
                       (grid.id, c.id, c.Parent.id)

            # all refinement grids should have parent 
            if grid.Level > 0:
                assert grid.Parent is not None and grid.Parent.id >= 0, \
                       'Grid %d, Level %d, Parent %d' % \
                       (grid.id, grid.Level, \
                        grid.Parent.id if grid.Parent is not None else -999)

            # parent index is consistent with the loaded dataset
            if grid.Level > 0:
                father_gid = father_list[grid.id]
                assert father_gid == grid.Parent.id, \
                       'Grid %d, Level %d, Parent_Found %d, Parent_Expect %d'%\
                       (grid.id, grid.Level, grid.Parent.id, father_gid)

            # edges between children and parent
            if len(grid.Children) > 0:
                assert_equal(grid.LeftEdge,  grid.Children[0].LeftEdge )
                assert_equal(grid.RightEdge, grid.Children[7].RightEdge)
        mylog.info('Check passed')
               

class GAMERDataset(Dataset):
    _index_class      = GAMERHierarchy
    _field_info_class = GAMERFieldInfo
    _handle           = None
    _debug            = False # debug mode for the GAMER frontend
    
    def __init__(self, filename,
                 dataset_type      = 'gamer',
                 storage_filename  = None,
                 particle_filename = None, 
                 units_override    = None,
                 unit_system       = "cgs"):

        if self._handle is not None: return

        self.fluid_types      += ('gamer',)
        self._handle           = HDF5FileHandler(filename)
        self.particle_filename = particle_filename

        if self.particle_filename is None:
            self._particle_handle = self._handle
        else:
            try:
                self._particle_handle = HDF5FileHandler(self.particle_filename)
            except:
                raise IOError(self.particle_filename)

        # currently GAMER only supports refinement by a factor of 2
        self.refine_by = 2

        Dataset.__init__(self, filename, dataset_type,
                         units_override = units_override,
                         unit_system    = unit_system)
        self.storage_filename = storage_filename
        
    def _set_code_unit_attributes(self):
        # GAMER does not assume any unit yet ...
        if len(self.units_override) == 0:
            mylog.warning("GAMER does not assume any unit ==> " +
                          "Use units_override to specify the units")

        for unit, cgs in [("length", "cm"), ("time", "s"), ("mass", "g")]:
            setdefaultattr(self, "%s_unit"%unit, self.quan(1.0, cgs))

            if len(self.units_override) == 0:
                mylog.warning("Assuming 1.0 = 1.0 %s", cgs)
        
    def _parse_parameter_file(self):
        self.unique_identifier = \
            int(os.stat(self.parameter_filename)[stat.ST_CTIME])

        # shortcuts for different simulation information
        KeyInfo   = self._handle['Info']['KeyInfo']
        InputPara = self._handle['Info']['InputPara']
        Makefile  = self._handle['Info']['Makefile']
        SymConst  = self._handle['Info']['SymConst']

        # simulation time and domain
        self.current_time      = KeyInfo['Time'][0]
        self.dimensionality    = 3  # always 3D
        self.domain_left_edge  = np.array([0.,0.,0.], dtype='float64')
        self.domain_right_edge = KeyInfo['BoxSize'].astype('float64')
        self.domain_dimensions = KeyInfo['NX0'].astype('int64')

        # periodicity
        periodic         = InputPara['Opt__BC_Flu'][0] == 0
        self.periodicity = (periodic,periodic,periodic)

        # cosmological parameters
        if Makefile['Comoving']:
            self.cosmological_simulation = 1
            self.current_redshift        = 1.0/self.current_time - 1.0
            self.omega_matter            = InputPara['OmegaM0'] 
            self.omega_lambda            = 1.0 - self.omega_matter
            self.hubble_constant         = 0.6955   # H0 is not set in GAMER
        else:
            self.cosmological_simulation = 0
            self.current_redshift        = 0.0
            self.omega_matter            = 0.0
            self.omega_lambda            = 0.0
            self.hubble_constant         = 0.0

        # code-specific parameters
        for t in KeyInfo, InputPara, Makefile, SymConst:
            for v in t.dtype.names: self.parameters[v] = t[v]

        # reset 'Model' to be more readable
        if KeyInfo['Model'] == 1:
            self.parameters['Model'] = 'Hydro'
        elif KeyInfo['Model'] == 2:
            self.parameters['Model'] = 'MHD'
        elif KeyInfo['Model'] == 3:
            self.parameters['Model'] = 'ELBDM'
        else:
            self.parameters['Model'] = 'Unknown'

        # make aliases to some frequently used variables
        if self.parameters['Model'] == 'Hydro' or \
           self.parameters['Model'] == 'MHD':
            self.gamma = self.parameters["Gamma"]
            self.mu    = self.parameters.get("mu",0.6) # mean molecular weight

    @classmethod
    def _is_valid(self, *args, **kwargs):
        try:
            # define a unique way to identify GAMER datasets
            f = HDF5FileHandler(args[0])
            if 'Info' in f['/'].keys() and 'KeyInfo' in f['/Info'].keys():
                return True
        except:
            pass
        return False



"""
GAMER-specific IO functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
from itertools import groupby

from yt.utilities.io_handler import \
    BaseIOHandler
from yt.utilities.logger import ytLogger as mylog


#-----------------------------------------------------------------------------
# GAMER shares a similar HDF5 format, and thus io.py as well, with FLASH
#-----------------------------------------------------------------------------


# group grids with consecutive indices together to improve the I/O performance
def grid_sequences(grids):
    for k, g in groupby( enumerate(grids), lambda i_x1:i_x1[0]-i_x1[1].id ):
        seq = list(v[1] for v in g)
        yield seq

class IOHandlerGAMER(BaseIOHandler):
    _particle_reader = False
    _dataset_type    = "gamer"

    def __init__(self, ds):
        super(IOHandlerGAMER, self).__init__(ds)
        self._handle      = ds._handle
        self._field_dtype = "float64" # fixed even when FLOAT8 is off

    def _read_particle_coords(self, chunks, ptf):
        pass

    def _read_particle_fields(self, chunks, ptf, selector):
        pass

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks) # generator --> list

        if any( (ftype != "gamer" for ftype, fname in fields) ):
            raise NotImplementedError

        rv = {}
        for field in fields: rv[field] = np.empty( size, dtype=self._field_dtype )

        ng = sum( len(c.objs) for c in chunks ) # c.objs is a list of grids
        mylog.debug( "Reading %s cells of %s fields in %s grids",
                     size, [f2 for f1, f2 in fields], ng )

        for field in fields:
            ds     = self._handle[ "/Data/%s" % field[1] ]
            offset = 0
            for chunk in chunks:
                for gs in grid_sequences(chunk.objs):
                    start = gs[ 0].id
                    end   = gs[-1].id + 1
                    data  = ds[start:end,:,:,:].transpose()
                    for i, g in enumerate(gs):
                        offset += g.select( selector, data[...,i], rv[field], offset )
        return rv

    def _read_chunk_data(self, chunk, fields):
        rv = {}
        if len(chunk.objs) == 0: return rv 

        for g in chunk.objs: rv[g.id] = {}

        for field in fields:
            ds = self._handle[ "/Data/%s" % field[1] ]

            for gs in grid_sequences(chunk.objs):
                start = gs[ 0].id
                end   = gs[-1].id + 1
                data  = ds[start:end,:,:,:].transpose()
                for i, g in enumerate(gs):
                    rv[g.id][field] = np.asarray( data[...,i], dtype=self._field_dtype )
        return rv



"""
YT is a package written primarily in Python designed to make the task of
running Enzo easier.  It contains facilities for creating Enzo data (currently
in prototype form) as well as runnning Enzo simulations, simulating the actions
of Enzo on various existing data, and analyzing output from Enzo in a
wide-variety of methods.

An ever-growing collection of documentation is also available at
http://yt-project.org/doc/ . Additionally, there is a
project site at http://yt-project.org/ with recipes, a wiki, a variety of
ways of peering into the version control, and a bug-reporting system.

YT is divided into several packages.

frontends
---------

This is where interfaces to codes are created.  Within each subdirectory of
yt/frontends/ there must exist the following files, even if empty:

* data_structures.py, where subclasses of AMRGridPatch, Dataset and
  AMRHierarchy are defined.
* io.py, where a subclass of IOHandler is defined.
* misc.py, where any miscellaneous functions or classes are defined.
* definitions.py, where any definitions specific to the frontend are
  defined.  (i.e., header formats, etc.)

visualization
-------------

This is where all visualization modules are stored.  This includes plot
collections, the volume rendering interface, and pixelization frontends.

data_objects
------------

All objects that handle data, processed or unprocessed, not explicitly
defined as visualization are located in here.  This includes the base
classes for data regions, covering grids, time series, and so on.  This
also includes derived fields and derived quantities.

analysis_modules
----------------

This is where all mechanisms for processing data live.  This includes
things like clump finding, halo profiling, halo finding, and so on.  This
is something of a catchall, but it serves as a level of greater
abstraction that simply data selection and modification.

gui
---

This is where all GUI components go.  Typically this will be some small
tool used for one or two things, which contains a launching mechanism on
the command line.

utilities
---------

All broadly useful code that doesn't clearly fit in one of the other
categories goes here.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

__version__ = "3.3.3"

# First module imports
import numpy as np # For modern purposes
import numpy # In case anyone wishes to use it by name

from yt.funcs import \
    iterable, \
    get_memory_usage, \
    print_tb, \
    rootonly, \
    insert_ipython, \
    get_pbar, \
    only_on_root, \
    is_root, \
    get_version_stack, \
    get_yt_supp, \
    get_yt_version, \
    parallel_profile, \
    enable_plugins, \
    memory_checker, \
    deprecated_class
from yt.utilities.logger import ytLogger as mylog

import yt.utilities.physical_constants as physical_constants
import yt.units as units
from yt.units.yt_array import \
    YTArray, \
    YTQuantity, \
    uconcatenate, \
    uintersect1d, \
    uunion1d, \
    loadtxt, \
    savetxt

from yt.fields.api import \
    field_plugins, \
    DerivedField, \
    FieldDetector, \
    FieldInfoContainer, \
    ValidateParameter, \
    ValidateDataField, \
    ValidateProperty, \
    ValidateSpatial, \
    ValidateGridType, \
    add_field, \
    derived_field

from yt.data_objects.api import \
    DatasetSeries, ImageArray, \
    particle_filter, add_particle_filter, \
    create_profile, Profile1D, Profile2D, Profile3D, \
    ParticleProfile

# For backwards compatibility
TimeSeriesData = deprecated_class(DatasetSeries)

from yt.frontends.api import _frontend_container
frontends = _frontend_container()

from yt.frontends.stream.api import \
    load_uniform_grid, load_amr_grids, \
    load_particles, load_hexahedral_mesh, load_octree, \
    hexahedral_connectivity, load_unstructured_mesh

from yt.frontends.ytdata.api import \
    save_as_dataset

# For backwards compatibility
GadgetDataset = frontends.gadget.GadgetDataset
GadgetStaticOutput = deprecated_class(GadgetDataset)
TipsyDataset = frontends.tipsy.TipsyDataset
TipsyStaticOutput = deprecated_class(TipsyDataset)

# Now individual component imports from the visualization API
from yt.visualization.api import \
    FixedResolutionBuffer, ObliqueFixedResolutionBuffer, \
    write_bitmap, write_image, \
    apply_colormap, scale_image, write_projection, \
    SlicePlot, AxisAlignedSlicePlot, OffAxisSlicePlot, \
    ProjectionPlot, OffAxisProjectionPlot, \
    show_colormaps, add_cmap, make_colormap, \
    ProfilePlot, PhasePlot, ParticlePhasePlot, \
    ParticleProjectionPlot, ParticleImageBuffer, ParticlePlot

from yt.visualization.volume_rendering.api import \
    volume_render, create_scene, ColorTransferFunction, TransferFunction, \
    off_axis_projection, interactive_render
import yt.visualization.volume_rendering.api as volume_rendering
#    TransferFunctionHelper, MultiVariateTransferFunction
#    off_axis_projection

from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_objects, enable_parallelism, communication_system

from yt.convenience import \
    load, simulation

from yt.testing import run_nose

# Import some helpful math utilities
from yt.utilities.math_utils import \
    ortho_find, quartiles, periodic_position

from yt.units.unit_systems import UnitSystem
from yt.units.unit_object import unit_system_registry

from yt.analysis_modules.list_modules import \
    amods



"""
Derived field base class.

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import contextlib
import inspect

from yt.extern.six import string_types, PY2
from yt.funcs import \
    ensure_list
from .field_exceptions import \
    NeedsGridType, \
    NeedsOriginalGrid, \
    NeedsDataField, \
    NeedsProperty, \
    NeedsParameter, \
    NeedsParameterValue, \
    FieldUnitsError
from .field_detector import \
    FieldDetector
from yt.units.unit_object import \
    Unit
import yt.units.dimensions as ytdims
from yt.utilities.exceptions import \
    YTFieldNotFound


def TranslationFunc(field_name):
    def _TranslationFunc(field, data):
        # We do a bunch of in-place modifications, so we will copy this.
        return data[field_name].copy()
    _TranslationFunc.alias_name = field_name
    return _TranslationFunc

def NullFunc(field, data):
    raise YTFieldNotFound(field.name)

class DerivedField(object):
    """
    This is the base class used to describe a cell-by-cell derived field.

    Parameters
    ----------

    name : str
       is the name of the field.
    function : callable
       A function handle that defines the field.  Should accept
       arguments (field, data)
    units : str
       A plain text string encoding the unit, or a query to a unit system of
       a dataset. Powers must be in python syntax (** instead of ^). If set
       to "auto" the units will be inferred from the units of the return
       value of the field function, and the dimensions keyword must also be
       set (see below).
    take_log : bool
       Describes whether the field should be logged
    validators : list
       A list of :class:`FieldValidator` objects
    particle_type : bool
       Is this a particle (1D) field?
    vector_field : bool
       Describes the dimensionality of the field.  Currently unused.
    display_field : bool
       Governs its appearance in the dropdowns in Reason
    not_in_all : bool
       Used for baryon fields from the data that are not in all the grids
    display_name : str
       A name used in the plots
    output_units : str
       For fields that exist on disk, which we may want to convert to other
       fields or that get aliased to themselves, we can specify a different
       desired output unit than the unit found on disk.
    dimensions : str or object from yt.units.dimensions
       The dimensions of the field, only needed if units="auto" and only used
       for error checking.
    """
    def __init__(self, name, function, units=None,
                 take_log=True, validators=None,
                 particle_type=False, vector_field=False, display_field=True,
                 not_in_all=False, display_name=None, output_units=None,
                 dimensions=None, ds=None):
        self.name = name
        self.take_log = take_log
        self.display_name = display_name
        self.not_in_all = not_in_all
        self.display_field = display_field
        self.particle_type = particle_type
        self.vector_field = vector_field
        self.ds = ds

        self._function = function

        if validators:
            self.validators = ensure_list(validators)
        else:
            self.validators = []

        # handle units
        if units is None:
            self.units = ''
        elif isinstance(units, string_types):
            if units.lower() == 'auto':
                if dimensions is None:
                    raise RuntimeError("To set units='auto', please specify the dimensions "
                                       "of the field with dimensions=<dimensions of field>!")
                self.units = None
            else:
                self.units = units
        elif isinstance(units, Unit):
            self.units = str(units)
        else:
            raise FieldUnitsError("Cannot handle units '%s' (type %s)." \
                                  "Please provide a string or Unit " \
                                  "object." % (units, type(units)) )
        if output_units is None:
            output_units = self.units
        self.output_units = output_units

        if isinstance(dimensions, string_types):
            dimensions = getattr(ytdims, dimensions)
        self.dimensions = dimensions

    def _copy_def(self):
        dd = {}
        dd['name'] = self.name
        dd['units'] = self.units
        dd['take_log'] = self.take_log
        dd['validators'] = list(self.validators)
        dd['particle_type'] = self.particle_type
        dd['vector_field'] = self.vector_field
        dd['display_field'] = True
        dd['not_in_all'] = self.not_in_all
        dd['display_name'] = self.display_name
        return dd

    def get_units(self):
        if self.ds is not None:
            u = Unit(self.units, registry=self.ds.unit_registry)
        else:
            u = Unit(self.units)
        return u.latex_representation()

    def get_projected_units(self):
        if self.ds is not None:
            u = Unit(self.units, registry=self.ds.unit_registry)
        else:
            u = Unit(self.units)
        return (u*Unit('cm')).latex_representation()

    def check_available(self, data):
        """
        This raises an exception of the appropriate type if the set of
        validation mechanisms are not met, and otherwise returns True.
        """
        for validator in self.validators:
            validator(data)
        # If we don't get an exception, we're good to go
        return True

    def get_dependencies(self, *args, **kwargs):
        """
        This returns a list of names of fields that this field depends on.
        """
        e = FieldDetector(*args, **kwargs)
        if self._function.__name__ == '<lambda>':
            e.requested.append(self.name)
        else:
            e[self.name]
        return e

    _unit_registry = None
    @contextlib.contextmanager
    def unit_registry(self, data):
        old_registry = self._unit_registry
        if hasattr(data, 'unit_registry'):
            ur = data.unit_registry
        elif hasattr(data, 'ds'):
            ur = data.ds.unit_registry
        else:
            ur = None
        self._unit_registry = ur
        yield
        self._unit_registry = old_registry

    def __call__(self, data):
        """ Return the value of the field in a given *data* object. """
        self.check_available(data)
        original_fields = data.keys() # Copy
        if self._function is NullFunc:
            raise RuntimeError(
                "Something has gone terribly wrong, _function is NullFunc " +
                "for %s" % (self.name,))
        with self.unit_registry(data):
            dd = self._function(self, data)
        for field_name in data.keys():
            if field_name not in original_fields:
                del data[field_name]
        return dd

    def get_source(self):
        """
        Return a string containing the source of the function (if possible.)
        """
        return inspect.getsource(self._function)

    def get_label(self, projected=False):
        """
        Return a data label for the given field, including units.
        """
        name = self.name[1]
        if self.display_name is not None:
            name = self.display_name

        # Start with the field name
        data_label = r"$\rm{%s}" % name

        # Grab the correct units
        if projected:
            raise NotImplementedError
        else:
            if self.ds is not None:
                units = Unit(self.units, registry=self.ds.unit_registry)
            else:
                units = Unit(self.units)
        # Add unit label
        if not units.is_dimensionless:
            data_label += r"\ \ (%s)" % (units.latex_representation())

        data_label += r"$"
        return data_label

    def __repr__(self):
        if PY2:
            func_name = self._function.func_name
        else:
            func_name = self._function.__name__
        if self._function == NullFunc:
            s = "On-Disk Field "
        elif func_name == "_TranslationFunc":
            s = "Alias Field for \"%s\" " % (self._function.alias_name,)
        else:
            s = "Derived Field "
        if isinstance(self.name, tuple):
            s += "(%s, %s): " % self.name
        else:
            s += "%s: " % (self.name)
        s += "(units: %s" % self.units
        if self.display_name is not None:
            s += ", display_name: '%s'" % (self.display_name)
        if self.particle_type:
            s += ", particle field"
        s += ")"
        return s

class FieldValidator(object):
    pass

class ValidateParameter(FieldValidator):
    def __init__(self, parameters, parameter_values=None):
        """
        This validator ensures that the dataset has a given parameter.

        If *parameter_values* is supplied, this will also ensure that the field
        is available for all permutations of the field parameter.
        """
        FieldValidator.__init__(self)
        self.parameters = ensure_list(parameters)
        self.parameter_values = parameter_values
    def __call__(self, data):
        doesnt_have = []
        if self.parameter_values is not None:
            if isinstance(data, FieldDetector):
                raise NeedsParameterValue(self.parameter_values)
        for p in self.parameters:
            if not data.has_field_parameter(p):
                doesnt_have.append(p)
        if len(doesnt_have) > 0:
            raise NeedsParameter(doesnt_have)
        return True

class ValidateDataField(FieldValidator):
    def __init__(self, field):
        """
        This validator ensures that the output file has a given data field stored
        in it.
        """
        FieldValidator.__init__(self)
        self.fields = ensure_list(field)
    def __call__(self, data):
        doesnt_have = []
        if isinstance(data, FieldDetector): return True
        for f in self.fields:
            if f not in data.index.field_list:
                doesnt_have.append(f)
        if len(doesnt_have) > 0:
            raise NeedsDataField(doesnt_have)
        return True

class ValidateProperty(FieldValidator):
    def __init__(self, prop):
        """
        This validator ensures that the data object has a given python attribute.
        """
        FieldValidator.__init__(self)
        self.prop = ensure_list(prop)
    def __call__(self, data):
        doesnt_have = []
        for p in self.prop:
            if not hasattr(data,p):
                doesnt_have.append(p)
        if len(doesnt_have) > 0:
            raise NeedsProperty(doesnt_have)
        return True

class ValidateSpatial(FieldValidator):
    def __init__(self, ghost_zones = 0, fields=None):
        """
        This validator ensures that the data handed to the field is of spatial
        nature -- that is to say, 3-D.
        """
        FieldValidator.__init__(self)
        self.ghost_zones = ghost_zones
        self.fields = fields
    def __call__(self, data):
        # When we say spatial information, we really mean
        # that it has a three-dimensional data structure
        #if isinstance(data, FieldDetector): return True
        if not getattr(data, '_spatial', False):
            raise NeedsGridType(self.ghost_zones,self.fields)
        if self.ghost_zones <= data._num_ghost_zones:
            return True
        raise NeedsGridType(self.ghost_zones,self.fields)

class ValidateGridType(FieldValidator):
    def __init__(self):
        """
        This validator ensures that the data handed to the field is an actual
        grid patch, not a covering grid of any kind.
        """
        FieldValidator.__init__(self)
    def __call__(self, data):
        # We need to make sure that it's an actual AMR grid
        if isinstance(data, FieldDetector): return True
        if getattr(data, "_type_name", None) == 'grid': return True
        raise NeedsOriginalGrid()



import numpy as np

from yt.testing import \
    assert_equal, \
    fake_amr_ds, \
    fake_particle_ds, \
    fake_random_ds

# This will test the "dataset access" method.

def test_box_creation():
    ds = fake_random_ds(32, length_unit=2)
    left_edge = ds.arr([0.2, 0.2, 0.2], 'cm')
    right_edge = ds.arr([0.6, 0.6, 0.6], 'cm')
    center = (left_edge + right_edge)/2

    boxes = [
        ds.box(left_edge, right_edge),
        ds.box(0.5*np.array(left_edge), 0.5*np.array(right_edge)),
        ds.box((0.5*left_edge).tolist(), (0.5*right_edge).tolist())
    ]

    region = ds.region(center, left_edge, right_edge)

    for b in boxes:
        assert_equal(b.left_edge, region.left_edge)
        assert_equal(b.right_edge, region.right_edge)
        assert_equal(b.center, region.center)

def test_region_from_d():
    ds = fake_amr_ds(fields=["density"])
    # We'll do a couple here

    # First, no string units
    reg1 = ds.r[0.2:0.3,0.4:0.6,:]
    reg2 = ds.region([0.25, 0.5, 0.5], [0.2, 0.4, 0.0], [0.3, 0.6, 1.0])
    yield assert_equal, reg1["density"], reg2["density"]

    # Now, string units in some -- 1.0 == cm
    reg1 = ds.r[(0.1, 'cm'):(0.5, 'cm'), :, (0.25, 'cm'): (0.35, 'cm')]
    reg2 = ds.region([0.3, 0.5, 0.3], [0.1, 0.0, 0.25], [0.5, 1.0, 0.35])
    yield assert_equal, reg1["density"], reg2["density"]

    # Now, string units in some -- 1.0 == cm
    reg1 = ds.r[(0.1, 'cm'):(0.5, 'cm'), :, 0.25:0.35]
    reg2 = ds.region([0.3, 0.5, 0.3], [0.1, 0.0, 0.25], [0.5, 1.0, 0.35])
    yield assert_equal, reg1["density"], reg2["density"]

    # And, lots of : usage!
    reg1 = ds.r[:, :, :]
    reg2 = ds.all_data()
    yield assert_equal, reg1["density"], reg2["density"]

def test_accessing_all_data():
    # This will test first that we can access all_data, and next that we can
    # access it multiple times and get the *same object*.
    ds = fake_amr_ds(fields=["density"])
    dd = ds.all_data()
    yield assert_equal, ds.r["density"], dd["density"]
    # Now let's assert that it's the same object
    rho = ds.r["density"]
    rho *= 2.0
    yield assert_equal, dd["density"]*2.0, ds.r["density"]
    yield assert_equal, dd["gas", "density"]*2.0, ds.r["gas", "density"]

def test_particle_counts():
    ds = fake_random_ds(16, particles=100)
    assert ds.particle_type_counts == {'io': 100}

    pds = fake_particle_ds(npart=128)
    assert pds.particle_type_counts == {'io': 128}
